//-----------------------------------------------------------------------------
// TDosMove ver 1.02
//
// Last updated at: 09/10/1998
//
// Component that allows you to move thourgh the controls in your app with
//  UP/DOWN arrows or ENTER key insted of using the old boring TAB.
//
// Code by: Liran Shahar
//          Israel
//          simpletech@ibm.net
//-----------------------------------------------------------------------------
// Modify Version 1.02h
// Designer : Danny Tzu
// E-Mail: deven_tzu.tw@yahoo.com.tw
// 廡ο~ץC Ь README.TXT
unit DosMove;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Grids, DBCtrls, ComCtrls;

type
  TMoveOptions = set of (moEnter,moUpDn);

  TDosMove = class(TComponent)
  private
    FActive,
    FActiveColor: Boolean;
    FOptions: TMoveOptions;
    FEditNoBeep: Boolean;
    FLastWasEdit: Boolean;
    FIsTabGrid: Boolean;
    FTabGrid: Boolean;
    FPageControl: TPageControl;
    FSelectPageOptions: TShiftState;
    FEnterColor,
    FOldColor: TColor;
    FOldControl: TControl;
    FNPClassNames: TStringList;  // Not Process ClassNames

    FOverWrite: Boolean; //޳
    FIsLoad: Boolean;    //޳
    Msg: TWMKey;         //޳

    FOwnerKeyDown: TKeyEvent;
    FOwnerKeyPress: TKeyPressEvent;
    FActiveControlChange: TNotifyEvent;

    procedure SetTabGrid(const Value: Boolean);
    procedure SetActive(const Value: boolean);
    procedure SetKeyEvent;
    procedure SetSelectPageOptions(const Value: TShiftState);
    procedure SetActiveColor(const Value: boolean);
    procedure SetControlColor(AControl: TControl);
    procedure SetNPClassNames(const Value: TStringList);

  protected
    procedure NewKeyDown(Sender : TObject;var Key : word;Shift : TShiftState);
    procedure NewKeyPress(Sender : Tobject;var Key : char);
    procedure ActiveControlChange(Sender: TObject);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    property Active: boolean read FActive write SetActive default False;
    property ActiveColor: boolean read FActiveColor write SetActiveColor default False;
    property Options: TMoveOptions read FOptions write FOptions default [moEnter,moUpDn];
    property EditNoBeep: boolean read FEditNoBeep write FEditNoBeep default True;
    property TabGrid: Boolean read FTabGrid write SetTabGrid default True;
    property PageControl: TPageControl read FPageControl write FPageControl;
    property SelectPageOptions: TShiftState read FSelectPageOptions write SetSelectPageOptions default [ssCtrl];
    property EnterColor: TColor read FEnterColor write FEnterColor default clWindow;
    property NPClassNames: TStringList read FNPClassNames write SetNPClassNames;
  end;

procedure Register;

implementation

type
  TDosMoveControl = Class(TControl)
  published
    property Color;
  end;

procedure Register;
begin
  RegisterComponents('HomeMade', [TDosMove]);
end;

constructor TDosMove.Create(AOwner : TComponent);
var
  Loop : integer;
begin
  // First check to see no other TDosMove exists on the form
  for Loop := 0 to AOwner.ComponentCount-1 do
    if AOwner.Components[Loop] is TDosMove then
      raise
        EInvalidOperation.Create('TDosMove can have only one instance per form');

  // Create component and set default properties
  inherited Create(AOwner);
  FActive := False;
  FOptions := [moEnter,moUpDn];
  FEditNoBeep := True;
  FIsTabGrid := False;
  FTabGrid := True;
  (AOwner as TForm).KeyPreview := True;
  FEnterColor := clWindow;

  FSelectPageOptions := [ssCtrl];
  FPageControl := nil;
  FOldControl := nil;
  FActiveControlChange := nil;

  FOverWrite := False; //޳
  FIsLoad := False;    //޳

  FNPClassNames := TStringList.Create;

  FNPClassNames.Add('TPageControl');
  FNPClassNames.Add('TTabControl');
  FNPClassNames.Add('TButton');
  FNPClassNames.Add('TDBGrid');
  FNPClassNames.Add('TBitBtn');
  FNPClassNames.Sorted := True;
end; // Create

destructor TDosMove.Destroy;
begin
  ActiveColor := False;
  FOldControl := nil;
  FNPClassNames.Free;

  inherited Destroy;
end;

procedure TDosMove.ActiveControlChange(Sender: TObject);
var AControl: TControl;
begin
  try
    if Assigned(Screen) then
    begin
      AControl := Screen.ActiveControl;
      if not (csDesigning in ComponentState) then
        if (AControl <> nil) and
           (AControl is TControl) then
          SetControlColor(AControl);
    end;
  finally
    if Assigned(FActiveControlChange) then
       FActiveControlChange(Sender);
  end;
end;

procedure TDosMove.NewKeyDown(Sender : TObject;var Key : word;
  Shift : TShiftState);
var
  cmActive: TWinControl;
  booDropDown: Boolean;
begin
  booDropDown := False;
  // Call owner OnKeyDown if it's assigned
  if Assigned(FOwnerKeyDown) then
    FOwnerKeyDown(Sender,Key,Shift);

  if FActive then
  begin
    if Key = vk_Insert then
      FOverWrite := not FOverWrite; //޳

    // true if last active control is TCustomEdit and above
    cmActive := (Owner as TForm).ActiveControl;

    FLastWasEdit := cmActive is TCustomEdit;

    if (FOptions <> []) and (Shift = []) then
    begin
      if cmActive is TCustomGrid then
      begin
        if (Key = VK_RETURN) and FTabGrid then
        begin
          Key := VK_TAB;
          FIsTabGrid := True;
        end;
      end else
        if (cmActive is TCustomComboBox) or
           (cmActive is TDBLookupComboBox) then
        begin
          // P_O_wgܿﶵ
          if (cmActive is TCustomComboBox) then
            booDropDown := TCustomComboBox(cmActive).DroppedDown;
          if (cmActive is TDBLookupComboBox) then
            booDropDown := TDBLookupComboBox(cmActive).ListVisible;

          if not booDropDown then
          begin
            if (Key = VK_DOWN) or (Key = VK_RETURN) then
            begin
              (Owner as TForm).Perform(WM_NEXTDLGCTL, 0, 0);
              Key := 0;
            end;
            if (Key = VK_UP) then
            begin
              (Owner as TForm).Perform(WM_NEXTDLGCTL, 1, 0);
              Key := 0;
            end;
          end;
        end else
        begin
          if (cmActive is TCustomListBox) then
          begin
            // V [ <- ] W@Ӥ
            if (Key = VK_LEFT) then
              (Owner as TForm).Perform(WM_NEXTDLGCTL, 1, 0);
            // V [ -> ] U@Ӥ
            if (Key = VK_RIGHT) then
              (Owner as TForm).Perform(WM_NEXTDLGCTL, 0, 0);
          end else
          begin
            // Handle the specials keys
            if ((Key = VK_DOWN) and (moUpDn in FOptions)) or
               ((Key = VK_RETURN) and (moEnter in FOptions)) then
              (Owner as TForm).Perform(WM_NEXTDLGCTL, 0, 0)
            else
              if (Key = VK_UP) and (moUpDn in FOptions) then
                (Owner as TForm).Perform(WM_NEXTDLGCTL, 1, 0);

            //޳
            if (FOverWrite) and
               (TCustomEdit((Owner as TForm).ActiveControl).SelLength = 0) and
               (TCustomEdit((Owner as TForm).ActiveControl).SelStart <
                TCustomEdit((Owner as TForm).ActiveControl).GetTextLen) and
               (Key <> vk_Return) and (Key <> vk_Up) and (Key <> vk_Down) and
               (Key <> vk_Left) and (Key <> vk_Right) and (Key <> vk_Insert) and
               (Key <> vk_Back) then
            begin
              //TCustomEdit((Owner as TForm).ActiveControl).SelLength := 1;
              if Msg.CharCode > 127 then
                 TCustomEdit((Owner as TForm).ActiveControl).SelLength := 2
              else
              begin
                TCustomEdit((Owner as TForm).ActiveControl).SelLength := 2;
                { if char on current Caret positon is a Chinese word }
                if Ord(TCustomEdit((Owner as TForm).ActiveControl).SelText[1]) > 127 then
                   TCustomEdit((Owner as TForm).ActiveControl).SelLength := 2
                else
                   TCustomEdit((Owner as TForm).ActiveControl).SelLength := 1;
                end;
            end;
          end;
        end;
    end else
    begin
      if Assigned(PageControl) and (FSelectPageOptions = Shift) then
      begin
        if (Key >= 49) and (Key <= 57) then
          PageControl.ActivePageIndex := (Key - 49);
        
        if (Key = 192) then
        begin
          // `
          if (PageControl.ActivePageIndex + 1)
             < PageControl.ControlCount then
            PageControl.ActivePageIndex := PageControl.ActivePageIndex + 1
          else
            PageControl.ActivePageIndex := 0;
        end;
      end;
    end; // if Options<>[] ...

  end; // if FActive ...

end; // NewKeyDown

//-----------------------------------------------------------------------------
procedure TDosMove.NewKeyPress(Sender : TObject;var Key : char);
begin
  // Call owner OnKeyPress if it's assigned
  if Assigned(FOwnerKeyPress) then
    FOwnerKeyPress(Sender, Key);

  // Handle 'Enter' key that makes Edits beep
  if FActive then
  begin
    if FEditNoBeep and
       FLastWasEdit and
       (Key = #13) then
      Key := #0;

    if FIsTabGrid then
    begin
      FIsTabGrid := False;
      Key := #0;
    end;
  end; // if FActive ...
  
end; // NewKeyPress

procedure TDosMove.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if (AComponent = PageControl) then
      PageControl := nil;
end;

procedure TDosMove.SetActive(const Value: boolean);
begin
  FActive := Value;
  SetKeyEvent;
end;

procedure TDosMove.SetActiveColor(const Value: boolean);
begin
  if (FActiveColor <> Value) then
  begin
    FActiveColor := Value;
    if FActiveColor then
    begin
      if Assigned(Screen) then
      begin
        FActiveControlChange := Screen.OnActiveControlChange;
        Screen.OnActiveControlChange := ActiveControlChange;
      end;
    end else
      Screen.OnActiveControlChange := FActiveControlChange;
  end;
end;

procedure TDosMove.SetKeyEvent;
begin
  // Intercept with OnKeyDown event and OnKeyPress event of 'Owner'

  // Run Time
  if not (csDesigning in ComponentState) then
  begin
    if FActive then
    begin
      // ]w
      FOwnerKeyDown := (Self.Owner as TForm).OnKeyDown;
      (Self.Owner as TForm).OnKeyDown := NewKeyDown;
      FOwnerKeyPress := (Self.Owner as TForm).OnKeyPress;
      (Self.Owner as TForm).OnKeyPress := NewKeyPress;
    end else
    begin
      // ٭
      (Self.Owner as TForm).OnKeyDown := FOwnerKeyDown;
      (Self.Owner as TForm).OnKeyPress := FOwnerKeyPress;
    end;
  end;
end;

procedure TDosMove.SetSelectPageOptions(const Value: TShiftState);
begin
  // @u@
  FSelectPageOptions := Value - FSelectPageOptions;
end;

procedure TDosMove.SetTabGrid(const Value: Boolean);
begin
  FTabGrid := Value;
end;

procedure TDosMove.SetControlColor(AControl: TControl);
  function CanProcessClassNames(FromClassName: String): Boolean;
  var i: Integer;
  begin
    FNPClassNames.Find(FromClassName, i);
    Result := (i > 0);
  end;
begin
  if Assigned(AControl) then
  begin
    if Assigned(FOldControl) then
      TDosMoveControl(FOldControl).Color := FOldColor;

    if CanProcessClassNames(AControl.ClassName) and FActiveColor then
    begin
      FOldControl := AControl;
      FOldColor := TDosMoveControl(AControl).Color;
      TDosMoveControl(AControl).Color := FEnterColor;
    end else
    begin
      FOldControl := nil;
    end;
  end;
end;

procedure TDosMove.SetNPClassNames(const Value: TStringList);
begin
  FNPClassNames.Assign(Value);
end;

end.
